"""cubicweb-tsfacets application package

This cube implements facets using postgresql text search vectors.
"""

import re
import string

from rql import RQLHelper

from cubicweb.server.querier import QuerierHelper, manual_build_descr
from cubicweb.server.sources.rql2sql import SQLGenerator


INVALID_TSQUERY_CHARS = string.punctuation + "\0"
SANITIZE_TRANSMAP = str.maketrans(
    INVALID_TSQUERY_CHARS, len(INVALID_TSQUERY_CHARS) * " "
)


class TSFacetsSQLGenerator(SQLGenerator):
    def __init__(self, schema, dbhelper, tsft_table, tsft_where):
        super().__init__(schema, dbhelper)
        self.tsft_table = tsft_table
        self.tsft_where = tsft_where

    def _solutions_sql(self, select, solutions, distinct, needalias):
        sqls = []
        for solution in solutions:
            self._state.reset(solution)
            # visit restriction subtree
            if select.where is not None:
                self._state.add_restriction(select.where.accept(self))
            sql = [self._selection_sql(select.selection, distinct, needalias)]
            main_variable = re.search(
                r"^SELECT(\sDISTINCT)?\s(?P<main_var>\w+\.\w+)", sql[0]
            )["main_var"]
            if self._state.restrictions:
                where_part = " AND ".join(self._state.restrictions)
                sql.append(
                    f"WHERE {where_part} AND {self.tsft_where} AND TSFT.eid = {main_variable}"
                )
            else:
                sql.append(f"WHERE {self.tsft_where} AND TSFT.eid = {main_variable}")
            self._state.merge_source_cbs(self._state._needs_source_cb)
            # add required tables
            assert len(self._state.actual_tables) == 1, self._state.actual_tables
            tables = self._state.tables_sql()
            if tables:
                tables += f", {self.tsft_table} as TSFT"
                sql.insert(1, f"FROM {tables}")
            else:
                sql.insert(1, f"FROM {self.tsft_table} as TSFT")
            sqls.append("\n".join(sql))
        if distinct:
            return "\nUNION\n".join(sqls)
        else:
            return "\nUNION ALL\n".join(sqls)


def convert_rql_to_sql(cnx, rql, rql_args=None, sql_generator=None):
    rqlhelper = RQLHelper(
        cnx.vreg.schema,
        special_relations={"eid": "uid", "has_text": "fti"},
        backend="postgres",
    )
    qhelper = QuerierHelper(cnx.repo, cnx.vreg.schema)
    union = rqlhelper.parse(rql)
    rqlhelper.compute_solutions(union)
    rqlhelper.simplify(union)
    plan = qhelper.plan_factory(union, rql_args, cnx)
    plan.preprocess(union)
    for select in union.children:
        select.solutions.sort(key=lambda x: list(x.items()))
    if sql_generator is None:
        dbhelper = cnx.repo.system_source.dbhelper
        sql_generator = SQLGenerator(cnx.vreg.schema, dbhelper)
    sql, args, _ = sql_generator.generate(union, rql_args)
    return sql, args


def build_rset_descr(cnx, rql, args, results):
    rqlhelper = RQLHelper(
        cnx.vreg.schema,
        special_relations={"eid": "uid", "has_text": "fti"},
        backend="postgres",
    )
    rqlst = rqlhelper.parse(rql)
    rqlhelper.compute_solutions(rqlst, {"eid": cnx.entity_type}, kwargs=args)
    return manual_build_descr(cnx, rqlst, args, results)


def convert_to_tsquery(search_string: str) -> str:
    """build a ready-to-use tsquery

    remove all invalid / potential ts_ operators and add
    ':*' at the end of each word to perform a prefixed search.
    """
    sanitized = search_string.translate(SANITIZE_TRANSMAP)
    return "&".join(f"{chunk}:*" for chunk in sanitized.split())


def includeme(config):
    pass



from cubicweb_tsfacets import convert_rql_to_sql
from cubicweb_tsfacets.views import build_mapping_table_name


def build_facet_table(cnx, facet_cls):
    table_name = facet_cls.table_name
    ts_config = cnx.vreg.config["text-search-config"]
    target_etype_sql_part = ",".join(
        (f"'{etype}'" for etype in facet_cls.target_etypes)
    )
    cnx.system_sql(
        f"""
        CREATE TEMPORARY TABLE {table_name}_entities AS (
            SELECT eid FROM entities WHERE type IN ({target_etype_sql_part})
        )
    """
    )
    from_parts = []
    vector_parts = []
    args = {}
    for (
        facet_key,
        request_facet_def,
    ) in facet_cls.key_names_to_rql_definition.items():
        sql, _args = convert_rql_to_sql(cnx, request_facet_def.rql_request)
        for key, value in _args.items():
            # here we want to be sure we don't have the same keys on all
            # generated SQL
            new_k = str(hash((facet_key, sql, key)))
            sql = sql.replace(f"%({key})s", f"%({new_k})s")
            args[new_k] = value
        if request_facet_def.needs_mapping:
            # here we need to create a mapping table:
            rset = cnx.execute(request_facet_def.rql_request)
            all_values = set()
            for _target, value in rset:
                all_values.add(value)

            mapping_table_name = build_mapping_table_name(facet_cls, facet_key)
            cnx.system_sql(f"DROP TABLE IF EXISTS {mapping_table_name}")
            cnx.system_sql(
                f"""
                CREATE TABLE
                    {mapping_table_name} (id serial PRIMARY KEY, value varchar)
                """
            )

            if all_values:
                insert = f"""
                INSERT INTO {mapping_table_name} (value)
                values {",".join(["(%s)"] * len(all_values))}
                """
                cnx.system_sql(insert, list(all_values))
            cnx.system_sql(f"CREATE INDEX ON {mapping_table_name}(value)")
            from_parts.append(
                f"""
                LEFT OUTER JOIN (
                    SELECT
                        _f{facet_key}_to_be_mapped.target_eid,
                        {mapping_table_name}.id
                    FROM
                        ({sql}) as _f{facet_key}_to_be_mapped(target_eid,value)
                        JOIN {mapping_table_name} ON (
                            {mapping_table_name}.value = _f{facet_key}_to_be_mapped.value
                        )
                ) as _f{facet_key}(target_eid,value)
                ON entities.eid = _f{facet_key}.target_eid
                """
            )  # noqa
        else:
            from_parts.append(
                f"""
                LEFT OUTER JOIN ({sql}) as _f{facet_key}(target_eid,value)
                ON entities.eid = _f{facet_key}.target_eid
                """
            )
        vector_parts.append(
            f"""
            ARRAY_AGG(
                DISTINCT CASE WHEN _f{facet_key}.value is NULL
                THEN ''  ELSE '{facet_key}.' || _f{facet_key}.value END
            )
            """
        )
    if facet_cls.text_search_indexation:
        sql, _args = convert_rql_to_sql(cnx, facet_cls.text_search_indexation)
        for key, value in _args.items():
            # here we want to be sure we don't have the same keys on all
            # generated SQL
            new_k = str(hash((facet_key, sql, key)))
            sql = sql.replace(f"%({key})s", f"%({new_k})s")
            args[new_k] = value
        fti_vector_from = f"""
        LEFT OUTER JOIN ({sql}) as _tstable(target_eid, text)
        ON entities.eid = _tstable.target_eid
        """
        if ts_config:
            fti_vector = (
                f"TO_TSVECTOR('{ts_config}', _tstable.text) as textsearchvector"
            )
        else:
            fti_vector = "TO_TSVECTOR(_tstable.text) as textsearchvector"
    else:
        fti_vector = "NULL as textsearchvector"
        fti_vector_from = ""
    cnx.system_sql(f"DROP TABLE IF EXISTS {table_name}")
    cnx.system_sql(
        f"""
            CREATE TABLE {table_name} AS (
                SELECT
                    entities.eid,
                    ARRAY_TO_TSVECTOR(
                        ARRAY_REMOVE(
                            {"||".join(vector_parts)},
                            ''
                        )
                    ) as facetvector,
                    {fti_vector}
                FROM
                    {table_name}_entities AS entities
                    {" ".join(from_parts)}
                    {fti_vector_from}
                GROUP BY entities.eid
                {", _tstable.text" if fti_vector_from else ""}
            )
        """,
        args,
    )
    cnx.system_sql(f"CREATE INDEX {table_name}_eid_idx ON {table_name}(eid)")
    if fti_vector_from:
        cnx.system_sql(
            f"CREATE INDEX {table_name}_ginidx ON {table_name} USING GIN(textsearchvector)"
        )
    cnx.system_sql(
        f"CREATE INDEX {table_name}_fti_ginidx ON {table_name} USING GIN(facetvector)"
    )



# -*- coding: utf-8 -*-
# copyright 2022-2023 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, Set, Tuple, Optional, Union
from typing_extensions import TypedDict

from cubicweb.appobject import AppObject
from cubicweb.rset import ResultSet
from cubicweb_tsfacets import (
    convert_rql_to_sql,
    build_rset_descr,
    convert_to_tsquery,
    TSFacetsSQLGenerator,
)


class FacetItemType(TypedDict):
    value: str
    count: str


class RQLRequestFacetDef:
    def __init__(
        self,
        rql_request: str,
        needs_mapping: bool = False,
    ):
        self.rql_request = rql_request
        self.needs_mapping = needs_mapping


class TSFacets(AppObject):
    """Base class for facets declaration.

    Each subclass of it allow to define a group of facets
    allowing to retrieve target entities. These entities don't have
    to be of the same type.

    Each subclass represent a specific postgresql table.

    :attr:`key_names_to_rql_definition`
        a dictionary linking each facet key with a RQLRequestFacetDef object.
        This object represents a RQL request with the information of if we
        need a mapping table for the value or not. We need a mapping if we
        want to index string with space or other characters like "'".

    :attr:`text_search_indexation`
        a RQL request returning a list of tuples:
        (target entity eid, text to index for full text search)
    """

    __registry__ = "tsfacets"
    __abstract__ = True

    key_names_to_rql_definition: Dict[str, RQLRequestFacetDef] = {}

    text_search_indexation: str = ""

    target_etypes: Set = set()

    table_name: Optional[str] = None

    def get_mapping_table_name(self, facet_key: str) -> str:
        return build_mapping_table_name(self, facet_key)

    def _build_tsvector_parts(self, selected_facets: Dict[str, List[str]]):
        tsvector_parts = []
        for facet_key in self.key_names_to_rql_definition:
            values = selected_facets.get(facet_key)
            if not values:
                continue
            for value in values:
                tsvector_parts.append(f"{facet_key}.{value.lower()}")
        return tsvector_parts

    def _build_sql_parts(
        self,
        rql: str = None,
        rql_args: Dict = None,
        selected_facets: Dict[str, List[str]] = None,
        text_search: str = None,
        with_row_nb: bool = False,
    ) -> Tuple[str, str, Dict[str, str]]:
        if rql:
            converted_sql, sql_args = convert_rql_to_sql(self._cw.cnx, rql, rql_args)
            if with_row_nb:
                converted_sql = f"""
                SELECT _TABLE_TO_COUNT.*, ROW_NUMBER() OVER () AS row_nb
                FROM ({converted_sql}) as _TABLE_TO_COUNT
                """
            sql_restriction_part = f"""
            JOIN ({converted_sql}) as _TABLE_FROM_RQL(target_eid)
            ON TSFT.eid = _TABLE_FROM_RQL.target_eid
            """
        else:
            sql_restriction_part = ""
            sql_args = {}
        if rql_args:
            sql_args.update(rql_args)

        ts_vector_parts = self._build_tsvector_parts(selected_facets or {})
        where_parts = []
        if ts_vector_parts:
            # here we use double ' because we will be inside a function.
            where_parts.append(
                f"""
                TSFT.facetvector @@ to_tsquery(
                    '{'&'.join(ts_vector_parts)}'
                )
            """
            )

        if text_search:
            ts_config = self._cw.vreg.config["text-search-config"]
            query = convert_to_tsquery(text_search)
            if ts_config:
                where_parts.append(
                    f"""
                    to_tsquery(
                        '{ts_config}',
                        '{query}'
                    ) @@ TSFT.textsearchvector
                """
                )
            else:
                where_parts.append(
                    f"""
                    to_tsquery('{query}')
                    @@ TSFT.textsearchvector
                """
                )

        if where_parts:
            where = f"WHERE {' AND '.join(where_parts)}"
        else:
            where = ""

        return sql_restriction_part, where, sql_args

    def get_facets_values_with_count(
        self,
        rql_restriction: str = None,
        rql_args: Dict = None,
        selected_facets: Dict[str, List[str]] = None,
        text_search: str = None,
        keep_selected: bool = False,
    ) -> Dict[str, List[FacetItemType]]:
        """
        Retrieve all facet values with target entities count for each of them.
        :param rql_restriction: A RQL with a single variable in the Any close corresponding
        to the target entities.
        :param rql_args: arguments of rql_restriction.
        :param selected_facets: all selected facets as Dict[facet_key, [facet values]].
        :param text_search: a string to filter results.
        :param keep_selected: if you want to remove count from already selected values.
        :return: a dictionary of the type Dict[facet_key, Dict[value, target entity count]]
        """
        selected_facets = selected_facets or {}
        sql_restriction_part, where_sql_part, sql_args = self._build_sql_parts(
            rql_restriction,
            rql_args,
            selected_facets,
            text_search,
        )
        crs = self._cw.cnx.system_sql(
            f"""
            SELECT
                -- Break encoded words to get `code` and `value`
                split_part(word, '.', 1) as code,
                split_part(word, '.', 2) as value,
                ndoc
            FROM
                -- Word-count on all qualified records
                ts_stat($ts_stat$
                    SELECT TSFT.facetvector
                    FROM {self.table_name} as TSFT
                    {sql_restriction_part}
                    {where_sql_part}
                $ts_stat$)
            ORDER BY code, ndoc DESC;
            """,
            sql_args,
        )

        all_facets: Dict[str, List[FacetItemType]] = {}
        for code, value, count in crs.fetchall():
            if code not in self.key_names_to_rql_definition:
                continue
            if not keep_selected and value in selected_facets.get(code, []):
                continue

            all_facets.setdefault(code, []).append(
                {
                    "value": value,
                    "count": count,
                }
            )

        return all_facets

    def get_target_entities_count(
        self,
        rql_restriction: str = None,
        rql_args: Dict = None,
        selected_facets: Dict[str, List[str]] = None,
        text_search: str = None,
    ) -> int:
        """
        Count target entities taking into account selected facets.
        :param rql_restriction: A RQL with a single variable in the Any close corresponding
        to the target entities.
        :param rql_args: arguments of rql_restriction.
        :param selected_facets: all selected facets as Dict[facet_key, [facet values]].
        :param text_search: a string to filter results.
        :return: how many entities correspond to these facets.
        """
        if rql_restriction and not (text_search or selected_facets):
            return len(self._cw.execute(rql_restriction, rql_args, build_descr=False))
        sql_restriction_part, where_sql_part, sql_args = self._build_sql_parts(
            rql_restriction, rql_args, selected_facets, text_search
        )
        crs = self._cw.cnx.system_sql(
            f"""
            SELECT COUNT(*)
            FROM {self.table_name} as TSFT
            {sql_restriction_part}
            {where_sql_part}
            """,
            sql_args,
        )
        try:
            return int(crs.fetchall()[0][0])
        except IndexError:
            return 0

    def get_target_entities_rset(
        self,
        base_rql: str,
        selected_facets: Dict[str, List[str]],
        rql_args: Dict = None,
        text_search: str = None,
    ) -> ResultSet:
        """
        Construct a rset corresponding to the selected facets.
        :param base_rql: a rql allowing to get target entities with all needed parameters.
        Target entity have to be the first variable in the Any clause.
        :param selected_facets: all selected facets as Dict[facet_key, [facet values]].
        :param rql_args: arguments for base_rql.
        :param text_search: a string to filter results.
        :return: a CubicWeb ResultSet.
        """
        if not (text_search or selected_facets):
            return self._cw.execute(base_rql, rql_args)
        rql_args = rql_args or {}
        ts_vector_parts = self._build_tsvector_parts(selected_facets or {})
        where_parts = []
        if ts_vector_parts:
            where_parts.append(
                f"""
                            TSFT.facetvector @@ to_tsquery(
                                '{'&'.join(ts_vector_parts)}'
                            )
                        """
            )
        if text_search:
            ts_config = self._cw.vreg.config["text-search-config"]
            query = convert_to_tsquery(text_search)
            if ts_config:
                where_parts.append(
                    f"""
                    to_tsquery(
                        '{ts_config}',
                        '{query}'
                    ) @@ TSFT.textsearchvector
                """
                )
            else:
                where_parts.append(
                    f"""
                    to_tsquery('{query}')
                    @@ TSFT.textsearchvector
                """
                )
        if where_parts:
            tsft_where = " AND ".join(where_parts)
        else:
            tsft_where = ""
        sql_generator = TSFacetsSQLGenerator(
            self._cw.cnx.vreg.schema,
            self._cw.cnx.repo.system_source.dbhelper,
            tsft_table=self.table_name,
            tsft_where=tsft_where,
        )
        sql, sql_args = convert_rql_to_sql(
            self._cw.cnx, base_rql, rql_args, sql_generator=sql_generator
        )

        crs = self._cw.cnx.system_sql(
            sql,
            {
                **sql_args,
                **rql_args,
            },
        )
        results = crs.fetchall()
        description = build_rset_descr(self._cw, base_rql, rql_args, results)
        rset = ResultSet(results, base_rql, rql_args, description)
        rset.req = self._cw
        return rset

    def get_id_to_value_mapping(self, facet_key: str) -> Union[Dict[str, str], None]:
        """
        This method retrieves the mapping id -> value,
        if we are in the case a mapping was needed.
        :param facet_key: the key for which we want the mapping.
        :return: a dictionnary containing the mapping, or
        or None if there is no mapping.
        """
        if not self.key_names_to_rql_definition[facet_key].needs_mapping:
            return None
        crs = self._cw.cnx.system_sql(
            f"""
            SELECT CAST(id as TEXT), value FROM {self.get_mapping_table_name(facet_key)}
            WHERE value is not NULL AND value <> ''
            """
        )
        return dict(crs.fetchall())

    def get_id_to_value(self, facet_key: str, f_id: Union[str, int]) -> str:
        """
        This method retrieves value for a given facet_key and id,
        if we are in the case a mapping was needed.
        :param facet_key: the key for which we want the mapping.
        :param f_id: the id for which we want the value.
        :return: the corresponding value.
        """
        if not self.key_names_to_rql_definition[facet_key].needs_mapping:
            return f_id
        crs = self._cw.cnx.system_sql(
            f"""
            SELECT value FROM {self.get_mapping_table_name(facet_key)} WHERE id = %(id)s
            """,
            {"id": int(f_id)},
        )
        return crs.fetchall()[0][0]


def build_mapping_table_name(facet_cls: TSFacets, facet_key: str):
    return f"{facet_cls.table_name}_{facet_key}_mapping"



from cubicweb.cwctl import CWCTL
from cubicweb.toolsutils import Command
from cubicweb.utils import admincnx

from cubicweb_tsfacets.importer import build_facet_table


@CWCTL.register
class GenerateFacetsTables(Command):
    """Generate TSFacets tables.


    <instance>
      the identifier of the instance
    """

    name = "generate-tsfacets-tables"
    arguments = "<instance id>"
    min_args = max_args = 1

    def run(self, args):
        appid = args.pop(0)
        with admincnx(appid) as cnx:
            for regid in cnx.vreg["tsfacets"]:
                for facet_cls in cnx.vreg["tsfacets"][regid]:
                    build_facet_table(cnx, facet_cls)
                    cnx.commit()
